// dependonwin.cpp Copyright (C) 2008 񁗕ҏW 

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  This Program is free software; you can redistribute it and/or modify     *
 *  it under the terms of the GNU General Public License as published by     *
 *  the Free Software Foundation; either version 2, or (at your option)      *
 *  any later version.                                                       *
 *                                                                           *
 *  This Program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             *
 *  GNU General Public License for more details.                             *
 *                                                                           *
 *  You should have received a copy of the GNU General Public License        *
 *  along with GNU Make; see the file COPYING.  If not, write to             *
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.    *
 *  http://www.gnu.org/copyleft/gpl.html                                     *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <process.h>
#include <windows.h>

#include "nnc4hdus.h"
#include "uusbd.h"

void
miliSleep ( int s )
{
  Sleep( s );
}

bool
UsbDevice::Open ( ushort vendor, ushort product )
{
  husb = OpenDevice( vendor, product, true );
  return husb != 0;
}

UsbDevice::~UsbDevice ()
{
  if ( husb != 0 )
    CloseDevice( husb );
}

bool UsbDevice::vendorRequest (
  ushort wLength, uchar* data, uchar bRequest, uchar v0, uchar v1, uchar v2, uchar v3 )
{
  return VendorRequest( husb, bRequest,
                        (ushort)(v0 + (v1 << 8)),
                        (ushort)(v2 + (v3 << 8)),
                        wLength, (char*)data );
}

class RecManagementData {
public:
  RecContext* rc;
  HANDLE pipeHandle;
  OVERLAPPED overlapped[NUM_CONTEXT];
  int nbTransferred[NUM_BUFFER];

  double recStart;
  double lastTime;
  double progress;
  double maxInterval;
  int over;
  int readCount;
  int writeCount;
  int rbInd;
  int wbInd;

  RecManagementData ( RecContext* _rc )
	: rc( _rc ),
     recStart( 0 ), lastTime( 0 ), progress( -1 ), maxInterval( 0 ), over( 0 ),
     readCount( 0 ), writeCount( 0 ), rbInd( 0 ), wbInd( 0 ) {
    for ( int i = 0; i < NUM_CONTEXT; i++ ) {
      overlapped[i].Offset = overlapped[i].OffsetHigh = 0;
      overlapped[i].hEvent = (void*)this;
    }
  }
};

VOID CALLBACK
readCompletionRoutine (
  DWORD dwErrorCode, 
  DWORD dwNumberOfBytesTransferred, 
  LPOVERLAPPED lpOverlapped )
{
  RecManagementData* rmd = (RecManagementData*)lpOverlapped->hEvent;
  RecContext* rc = rmd->rc;
  double dt = getTime();
  if ( rmd->progress < 0 ) {
    rmd->recStart = dt;
  } else {
    double interval = dt - rmd->lastTime;
    if ( interval > rmd->maxInterval )
      rmd->maxInterval = interval;
  }
  rmd->lastTime = dt;
  rmd->progress = dt - rmd->recStart;
  if ( rmd->progress >= rc->time )
    interrupted = true;

  if ( interrupted ) {
    rmd->over++;
  } else {
    lpOverlapped->Offset = lpOverlapped->OffsetHigh = 0;
    int i = rmd->rbInd + NUM_CONTEXT;
    if ( i >= NUM_BUFFER )
      i -= NUM_BUFFER;
    ReadFileEx( rmd->pipeHandle, rc->recvBuf + TRANSFER_SIZE*i + BUFFER_ADDITION,
                TRANSFER_SIZE, lpOverlapped, readCompletionRoutine );
  }
  rmd->readCount++;
  rmd->nbTransferred[rmd->rbInd++] = dwNumberOfBytesTransferred;
  if ( rmd->rbInd >= NUM_BUFFER )
    rmd->rbInd = 0;
}

static unsigned int __stdcall
_rec ( void* context )
{
  SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL );
  
  RecManagementData* rmd = (RecManagementData*)context;
  RecContext* rc = rmd->rc;
  HANDLE pipeHandle = OpenPipe( rc->usb.husb, 0, rc->tunerID );
  rmd->pipeHandle = pipeHandle;

  for ( int i = 0; i < NUM_CONTEXT; i++ ) {
    ReadFileEx( pipeHandle, rc->recvBuf + TRANSFER_SIZE*i + BUFFER_ADDITION,
                TRANSFER_SIZE, rmd->overlapped+i, readCompletionRoutine );
  }
  while ( rmd->over < NUM_CONTEXT && SleepEx( 5000, TRUE ) )
    ;
  rmd->over = NUM_CONTEXT;
  CloseHandle( pipeHandle );
  _endthreadex( 0 );
  return 0;
}

void
rec ( RecContext* rc )
{
  SetPriorityClass( GetCurrentProcess(), HIGH_PRIORITY_CLASS );

  RecManagementData rmd( rc );
  unsigned threadID;
  HANDLE hThread = (HANDLE)_beginthreadex( NULL, 0, _rec, &rmd, 0, &threadID );
  int maxd = 0;
  do {
    Sleep( 10 );
    for (;;) {
      int d = rmd.readCount - rmd.writeCount;
      if ( d <= 0 )
        break;
      rc->syncOut( rc->recvBuf + BUFFER_ADDITION + TRANSFER_SIZE*rmd.wbInd,
                   rmd.nbTransferred[rmd.wbInd] );
      rmd.writeCount++;
      rmd.wbInd++;
      if ( rmd.wbInd >= NUM_BUFFER )
        rmd.wbInd = 0;
      if ( maxd < d )
        maxd = d;

      fprintf( stderr, "\rprogress  = %1.0fs / %1.0fs [ %4.3f : %02d ] ",
               rmd.progress, rc->time, rmd.maxInterval, maxd );
    }
  } while ( rmd.over < NUM_CONTEXT );
  WaitForSingleObject( hThread, INFINITE );
  CloseHandle( hThread );
  fputc( '\n', stderr );

  SetPriorityClass( GetCurrentProcess(), NORMAL_PRIORITY_CLASS );
}
