// tp4hdp2.cpp Copyright (C) 2008 񁗕ҏW 

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  This Program is free software; you can redistribute it and/or modify     *
 *  it under the terms of the GNU General Public License as published by     *
 *  the Free Software Foundation; either version 2, or (at your option)      *
 *  any later version.                                                       *
 *                                                                           *
 *  This Program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             *
 *  GNU General Public License for more details.                             *
 *                                                                           *
 *  You should have received a copy of the GNU General Public License        *
 *  along with GNU Make; see the file COPYING.  If not, write to             *
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.    *
 *  http://www.gnu.org/copyleft/gpl.html                                     *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>

#include "nnc4hdus.h"
#include "decrypt.h"

#define APP_NAME "tp4hdp2"
#define VERSION  "ver 0.0.0"

#define VENDOR_ID  (0x3275)
#define PRODUCT_ID (0x6111)

#define SA_ASIE5606_2     (0x5E)
#define SA_ASIE5606_NEW_2 (0x4E)
#define SA_TUNER_2        (0x30)

uchar  syncByte = 0x47;
uchar  sa_asie5606 = SA_ASIE5606;
uchar  sa_tuner    = SA_TUNER;
uchar  sa_asie5606_05 = 0x91;
ushort pidMaskStart = 0x1FFF;
ushort pidMaskEnd   = 0x1FFF;
uchar  key[0x10];

bool interrupted = false;
bool verbose = false;

static void sig_handler( int sig )
{
  fprintf( stderr,  "interrupted\n" );
  interrupted = true;
  signal( sig, sig_handler );
}

double
getTime ()
{
  struct timeval tv;
  gettimeofday( &tv, NULL );
  return tv.tv_sec + tv.tv_usec * (1.0/1000000);
}

void
dumpdata ( uchar* data, int size )
{
  int i = 0;
  while ( size > 0 ) {
    int i4 = i & 0xf;
    if ( i4 == 0 )
      fprintf( stderr, " %04X ", i );
    fprintf( stderr, "%02X", *data++ );
    fputc( i4 == 0xf? '\n': ' ', stderr );
    i++;
    size--;
  }
  if ( (i & 0xf) != 0x0 )
    fputc( '\n', stderr );
}

bool
TSPointer::isPacketTop ( uchar* p, int n )
{
  for ( int i = 0; i < n; i++ ) {
    if ( *p != syncByte )
      return false;
    p += PACKET_LENGTH;
  }
  return true;
}

bool
TSPointer::resync ( TSPointer& rem )
{
  for (;; len-- ) {
    if ( len <= BUFFER_ADDITION ) {
      rem = *this;
      return false;
    }
    if ( isPacketTop( ptr, NUM_CHECKPACKET ) ) {
      return true;
    }
    ptr++;
  }
}

bool
RecContext::out ( uchar* data, int len )
{
  if ( fwrite( data, len, 1, output ) == 1 ) 
    return false;
  return interrupted = true;
}

void
RecContext::syncOut ( uchar* buf, ulong size )
{
  TSPointer tsp( buf, size );
  if ( remainder.len > 0 ) {
    tsp.ptr -= remainder.len;
    tsp.len += remainder.len;
    if ( tsp.ptr != remainder.ptr )
      for ( int i = remainder.len - 1; i > 0; i-- )
        tsp.ptr[i] = remainder.ptr[i];
  }
  if ( !lock ) {
    if ( !tsp.resync( remainder ) )
      return;
    lock = true;
  }
  for (;;) {
    uchar* const top = tsp.ptr;
    int l = 0;
    for (;; tsp.len -= PACKET_LENGTH ) {
      if ( tsp.len <= PACKET_LENGTH ) {
        remainder = tsp;
        if ( l > 0 )
          out( top, l );
        return;
      }
      if ( tsp.ptr[PACKET_LENGTH] != syncByte ) {
        if ( l > 0 && out( top, l ) )
          return;
        tsp.ptr++;
        tsp.len--;
        if ( !tsp.resync( remainder ) ) {
          lock = false;
          return;
        }
        break;
      }
      tsp.ptr[0] = 0x47;
      packetDecrypt( tsp.ptr, tsp.ptr, 1 );
      tsp.ptr += PACKET_LENGTH;
      l += PACKET_LENGTH;
    }
  }
}

class HDUS : public RecContext {
  HDUS ( const HDUS & );
  uchar vrBuf[ 0x100 ];
public:
  int freq;
  const double startTime;
  double lastTime;
  bool first;
  const char* fileName;

  HDUS () : freq( 0 ), startTime( getTime() ), lastTime( 0 ), first( true ), fileName( "test.ts2" ) {}
  ~HDUS () {
    if ( output )
      fclose( output );
  }

  bool vendorRequest (
    ushort wLength, uchar* data, uchar bRequest,
    uchar v0 = 0, uchar v1 = 0, uchar v2 = 0, uchar v3 = 0 );

  void tunerPowerOn ();
  void asie5606PowerOn ();
  void powerOff ();

  void sendTuner ( uchar* data, int len );

  void initTuner ();
  void setFreq ( ulong freq );

  void dump ();
};

bool
HDUS::vendorRequest ( ushort wLength, uchar* data, uchar bRequest, uchar v0, uchar v1, uchar v2, uchar v3 )
{
#ifndef WITHOUT_USB
  bool r = usb.vendorRequest( wLength, data, bRequest, v0, v1, v2, v3 );
#else
  bool r = true;
#endif
  double currentTime = getTime();
  lastTime = currentTime - lastTime;
  if ( first || lastTime < 0.0 )
    lastTime = 0.0;
  if ( verbose ) {
    fprintf( stderr, "vendor request : %02X %02X %02X %02X %02X %04X : %.3f / +%.3f : %d\n",
             bRequest, v0, v1, v2, v3, wLength, currentTime - startTime, lastTime, r? 1: 0 );
    dumpdata( data, wLength );
  }
  lastTime = currentTime;
  first = false;
  return r;
}

void
HDUS::tunerPowerOn () // `[i[d͋LĂ 
{
  vendorRequest( 1, vrBuf, 0x08, 0x08, 0x08 );
  miliSleep( 50 );
  vendorRequest( 1, vrBuf, 0x08, 0x00, 0x08 );
  miliSleep( 150 );
  vendorRequest( 1, vrBuf, 0x08, 0xFF, 0x11 );
  miliSleep( 50 );
  vendorRequest( 1, vrBuf, 0x08, 0x00, 0xD5 );
  miliSleep( 50 );
  vendorRequest( 1, vrBuf, 0x08, 0xFF, 0xD5 );
  miliSleep( 100 );
}

void
HDUS::asie5606PowerOn ()
{
  uchar v = sa_tuner == SA_TUNER? 0x11: 0x80;
  vendorRequest( 1, vrBuf, 0x08, 0xFF, v ); // bit4ASIE5606bit0̓J[h肩H 
  miliSleep( 50 );
  vendorRequest( 1, vrBuf, 0x08, 0x00, v );
  miliSleep( 50 );
  vendorRequest( 1, vrBuf, 0x08, 0xFF, v );
  miliSleep( 100 );
}

void
HDUS::powerOff () // `[i[d͋LĂ 
{
  vendorRequest( 1, vrBuf, 0x08, 0x00, 0xD4 );
  vendorRequest( 1, vrBuf, 0x08, 0x08, 0x08 );
  miliSleep( 50 );
}

uchar tunerAddr[ 0x5e ] = {
  0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
  0x0C, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
  0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
  0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
  0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
  0x30, 0x31, 0x32, 0x34, 0x38, 0x39, 0x3A, 0x3B,
  0x3C, 0x3D, 0x3E, 0x3F, 0x40, 0x41, 0x42, 0x43,
  0x44, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C,
  0x50, 0x51, 0x52, 0x54, 0x55, 0x56, 0x57, 0x58,
  0x5C, 0x5D, 0x5E, 0x5F, 0x70, 0x71, 0x72, 0x75,
  0x76, 0x77, 0x7C, 0x7D, 0xBA, 0xBB, 0xBC, 0xC2,
  0xC7, 0xE4, 0xE6, 0xE9, 0xEC, 0xEF
};

uchar tunerData[ 0x5e ] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x21, 0x10, 0x03, 0xE0, 0x42, 0x09, 0x20, 
  0x29, 0x13, 0x29, 0x13, 0x2A, 0xAA, 0xAA, 0xA8,
  0x00, 0xFF, 0x80, 0x4C, 0x4C, 0x80, 0x00, 0x0C, 
  0x60, 0x6B, 0x40, 0x40, 0xFF, 0x00, 0xFF, 0x00,
  0x20, 0x0F, 0x84, 0x0F, 0x00, 0x0F, 0x20, 0x21, 
  0x3F, 0x10, 0x08, 0x0C, 0x0C, 0x00, 0x00, 0x4F,
  0xFF, 0x20, 0x00, 0x90, 0xE6, 0x02, 0x54, 0x00, 
  0x04, 0x58, 0x20, 0x57, 0xF1, 0x20, 0x70, 0x60,
  0x50, 0x00, 0x01, 0x80, 0x18, 0x00, 0x00, 0x22, 
  0x00, 0x01, 0x00, 0x52, 0x00, 0x00, 0x00, 0x10,
  0x00, 0x1A, 0x2F, 0x08, 0x00, 0x00
};

void
HDUS::sendTuner ( uchar* data, int len )
{
  data[0] = 0xFE; data[1] = sa_tuner == SA_TUNER? 0xC0: 0xC4;
  uchar offset = 0;
  for ( ; len >= 3; len -= 3 ) {
    vendorRequest( 4, vrBuf, 0x0D, offset, data[0], data[1], data[2] );
    data += 3;
    offset += 3;
  }
  switch ( len ) {
  case 2: vendorRequest( 3, vrBuf, 0x0D, offset, data[0], data[1] ); offset += 2; break;
  case 1: vendorRequest( 2, vrBuf, 0x0D, offset, data[0]          ); offset += 1; break;
  }
  vendorRequest( offset+1, vrBuf, 0x0E, sa_tuner );
}

void
HDUS::initTuner ()
{
  uchar data[4];
  data[2] = 0xFF; sendTuner( data, 3 );

  for ( unsigned int i = 0; i < sizeof( tunerAddr ); i++ ) {
    vendorRequest( 2, vrBuf, 0x03, sa_tuner, tunerAddr[i], tunerData[i] );
  }

  vendorRequest( 2, vrBuf, 0x02, sa_tuner, 0x9D );
  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x1C, (vrBuf[1] & 0xeb) | 0x28 );

  data[2] = 0x2C; data[3] = 0x44; sendTuner( data, 4 );
  data[2] = 0x4D; data[3] = 0x40; sendTuner( data, 4 );
  data[2] = 0x7F; data[3] = 0x02; sendTuner( data, 4 );
  data[2] = 0x9A; data[3] = 0x52; sendTuner( data, 4 );
  data[2] = 0x48; data[3] = 0x5A; sendTuner( data, 4 );
  data[2] = 0x76; data[3] = 0x1A; sendTuner( data, 4 );
  data[2] = 0x6A; data[3] = 0x48; sendTuner( data, 4 );
  data[2] = 0x64; data[3] = 0x28; sendTuner( data, 4 );
  data[2] = 0x66; data[3] = 0xE6; sendTuner( data, 4 );
  data[2] = 0x35; data[3] = 0x11; sendTuner( data, 4 );
  data[2] = 0x7E; data[3] = 0x01; sendTuner( data, 4 );
  data[2] = 0x0B; data[3] = 0x99; sendTuner( data, 4 );
  data[2] = 0x0C; data[3] = 0x00; sendTuner( data, 4 );
  data[2] = 0x10; data[3] = 0x40; sendTuner( data, 4 );
  if ( sa_tuner != SA_TUNER ) {
    data[3] = 0x00; sendTuner( data, 4 );
  }
  data[2] = 0x12; data[3] = 0xCA; sendTuner( data, 4 );
  data[2] = 0x16; data[3] = 0x90; sendTuner( data, 4 );
  data[2] = 0x32; data[3] = 0x36; sendTuner( data, 4 );
  data[2] = 0xD8; data[3] = 0x18; sendTuner( data, 4 );
  data[2] = 0x05; data[3] = 0x01; sendTuner( data, 4 );
}

void
HDUS::setFreq ( ulong freq )
{
  freq = (freq << 6) / 1000;

  vendorRequest( 2, vrBuf, 0x02, sa_tuner, 0x9D );
  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x1C, (vrBuf[1] & 0xfb) | 0x38 );

  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x25, 0x00 );
  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x23, 0x4D );

  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x1E, 0xAA );
  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x1F, 0xA8 );
  {
    uchar data[4];
    data[2] = 0x11; data[3] = 0x00; sendTuner( data, 4 );
    data[2] = 0x13; data[3] = 0x15; sendTuner( data, 4 );

    data[2] = 0x14; data[3] = (uchar)freq;        sendTuner( data, 4 );
    data[2] = 0x15; data[3] = (uchar)(freq >> 8); sendTuner( data, 4 );

    data[2] = 0x11; data[3] = 0x02; sendTuner( data, 4 );
  }
  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x1E, 0xA2 );
  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x1F, 0x08 );

  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x01, 0x40 );
  vendorRequest( 2, vrBuf, 0x03, sa_tuner, 0x23, 0x4C );
}

void
HDUS::dump ()
{
  uchar tunerID = sa_tuner == SA_TUNER? 0: 1;

  vendorRequest( 0x3A, vrBuf, 0x0C );
  vendorRequest( 3, vrBuf, 0x17, 0x02 );
  vendorRequest( 1, vrBuf, 0x08, 0x07, 0xFB );
  vendorRequest( 1, vrBuf, 0x10, 0x00, 0x02 );
  powerOff();

  vendorRequest( 3, vrBuf, 0x17, 0x02 );
  tunerPowerOn();
  initTuner();

  asie5606PowerOn();
  vendorRequest( 2, vrBuf, 0x02, sa_asie5606, 0x09 );
  if ( vrBuf[0] != 1 ) {
    sa_asie5606 += (SA_ASIE5606_NEW - SA_ASIE5606);
    vendorRequest( 2, vrBuf, 0x02, sa_asie5606, 0x09 );
#ifndef WITHOUT_USB
    if ( vrBuf[0] != 1 )
      goto _abort;
#endif
  }
  if ( (vrBuf[1] & 7) == 6 ) {
    sa_asie5606_05 = 0x80;
  }

  // 8000.b6ǂݏo vrBuf[1]1̏ꍇ̓nCXs[h(USB2)A0̏ꍇ̓tXs[h(USB1)ł 
  // ASV5211̓nCXs[h 
  vendorRequest( 1, vrBuf, 0x0A );
  miliSleep( 30 );
  // ASIEÍ[hݒ 
  vendorRequest( 2, vrBuf, 0x03, sa_asie5606, 0x05, sa_asie5606_05 );

  setFreq( (ulong)freq );
  vendorRequest( 1, vrBuf, 0x09, tunerID, 1 );

  // pidtB^ݒ 
  {
    uchar offset = sa_tuner == SA_TUNER? 0x00: 0x80;
    vendorRequest( 2, vrBuf, 0x04, 0x40+offset );
    vendorRequest( 2, vrBuf, 0x05, 0x40+offset, vrBuf[1] | 0x0F );
    vendorRequest( 3, vrBuf, 0x05, 0x41+offset, (uchar)(pidMaskStart >> 8), (uchar)pidMaskStart );
    vendorRequest( 3, vrBuf, 0x05, 0x43+offset, (uchar)(pidMaskEnd   >> 8), (uchar)pidMaskEnd );
  }
  // ASIE5606ÍL[ݒ 
  for ( uchar i = 0x0; i < 0x10; i++ )
    vendorRequest( 2, vrBuf, 0x03, sa_asie5606, i+0x10, key[i] );

  // ASV5211bulk]͂߂炷 
  vendorRequest( 1, vrBuf, 0x06, tunerID );

#ifndef WITHOUT_USB
  rec( this );
#endif

  miliSleep( 100 );

  // ASV5211bulk]߂炷 
  vendorRequest( 1, vrBuf, 0x07, tunerID ); 

  vendorRequest( 1, vrBuf, 0x09, tunerID, 1 );
  vendorRequest( 2, vrBuf, 0x03, sa_asie5606, 0x80, 0 );
  vendorRequest( 2, vrBuf, 0x03, sa_asie5606, 0x04, 0 );
  vendorRequest( 2, vrBuf, 0x03, sa_asie5606, 0x00, 0 );

_abort:
  miliSleep( 50 );
  powerOff();
}

int
main ( int argc, char** argv )
{
  if ( argc < 2 ) {
    fputs( APP_NAME " " VERSION "\n\n", stdout );
    fputs( "usage: " APP_NAME " @`l [] [t@C] [IvV]\n", stdout );
    fputs( "  `l : nfW\"@13\"`\"@62\"BCATVgϊ\"@C13\"`\"@C63\"B\n", stdout );
    fputs( "  Ԃ̂P : mmwB ( ŝƂ͕b ) \n", stdout );
    fputs( "  Ԃ̂Q : hh:mm-hh:mm ^Jn-IwB\n", stdout );
    fputs( "  t@C : t@Cȗ\"test.ts2\"B\n", stdout );
    fputs( "  IvV : -t1         `[i[#1gpB\n", stdout );
    fputs( "             : -v          vendor requestʐMOWG[ɏóB\n", stdout );
    fputs( "             : -mXXXX:XXXX pid}XNJn-IlݒB\n", stdout );
    fputc( '\n', stdout );
    fputs( "  ex: " APP_NAME " @27 1 : t@C\"test.ts2\"ch271^\n", stdout );
    return 0;
  }
  HDUS hdus;
  int sleepTime = 0;
  char* ch = 0;
  for ( int i = 1; i < argc; i++ ) {
    char *str = argv[i];
    unsigned int num[4];
    if ( str[0] == '@' ) {
      if ( (str[1] | 0x20) == 'c' ) {
        if ( sscanf( str+2, "%d", num ) == 1 && num[0] >= 13 && num[0] <= 63 ) {
          if ( num[0] >= 23 ) {
            if ( num[0] >= 24 && num[0] <= 27 ) {
              hdus.freq = 167000 + (num[0]-13)*6000;
            } else {
              hdus.freq = 165000 + (num[0]-13)*6000;
            }
          } else {
            if ( num[0] == 22 ) {
              hdus.freq = 113000 + (num[0]-13)*6000;
            } else {
              hdus.freq = 111000 + (num[0]-13)*6000;
            }
          }
          ch = str+1;
        }
      } else {
        if ( sscanf( str+1, "%d", num ) == 1 && num[0] >= 13 && num[0] <= 62 ) {
          hdus.freq = 473000 + (num[0]-13)*6000;
          ch = str+1;
        }
      }
    } else
    if ( str[0] == '-' ) {
      switch ( str[1] | 0x20 ) {
      case 'm':
        if ( sscanf( str+2, "%x:%x", num, num+1 ) == 2 &&
             num[0] <= 0x1FFF && num[1] <= 0x1FFF && num[0] <= num[1] ) {
          pidMaskStart = (ushort)num[0];
          pidMaskEnd   = (ushort)num[1];
        }
        break;
      case 't':
        if ( sscanf( str+2, "%d", num ) == 1 && num[0] == 1 ) {
          hdus.tunerID = 1;
          sa_tuner = SA_TUNER_2;
          sa_asie5606 += (SA_ASIE5606_2 - SA_ASIE5606);
        }
        break;
      case 'v':
        verbose = true;
        break;
      }
    } else
    if ( sscanf( str, "%u:%u-%u:%u", num, num+1, num+2, num+3 ) == 4 &&
         num[0] < 24 && num[1] < 60 && num[2] < 24 && num[3] < 60 ) {
      num[0] = num[0]*60 + num[1];
      num[2] = num[2]*60 + num[3];
      num[2] -= num[0];
      if ( (int)(num[2]) < 0 )
        num[2] += (60*24);
      hdus.time = num[2]*60.0;

      time_t t = time( 0 );
      struct tm* tmp = localtime( &t );
      sleepTime = (int)(num[0] - (tmp->tm_hour*60 + tmp->tm_min));
      if ( sleepTime < 0 )
        sleepTime += (60*24);
      sleepTime = sleepTime*60 - tmp->tm_sec - 2;
      fprintf( stderr, "sleep %ds\n", sleepTime );
    } else {
      char* err;
      ulong ul = strtoul( str, &err, 10 );
      if ( str < err ) {
        hdus.time = (double)ul;
        if ( (err[0] | 0x20) != 's' )
          hdus.time *= 60.0;
      } else {
        hdus.fileName = str;
      }
    }
  }
  if ( hdus.freq != 0 ) {
    if ( sleepTime > 0 ) {
      miliSleep( sleepTime * 1000 );
    }
#ifndef WITHOUT_USB
    if ( !hdus.usb.Open( VENDOR_ID, PRODUCT_ID ) ) {
#else
    if ( false ) {
#endif
      fputs( "no drivers for " APP_NAME, stderr );
    } else
    if ( !hdus.allocRecvBuf() ) {
      fputs( "not enough memory.", stderr );
    } else {
      hdus.output = fopen( hdus.fileName, "wb" );
      if ( hdus.output ) {
        fprintf( stderr, "Output = %s\n", hdus.fileName );
        fprintf( stderr, "ch = %s / Freq = %dKHz\n\n", ch, hdus.freq );
        if( signal( SIGINT, sig_handler ) == SIG_IGN )
          signal( SIGINT, SIG_IGN );
        hdus.dump();
      }
    }
  }
  return 0;
}
